@@ -1,5 +1,6 @@
 """Tools for performing image registration."""
 import matplotlib.pyplot as plt
+import matplotlib.cm
 import matplotlib.colors
 import numpy as np
 import os
@@ -9,7 +10,7 @@ import subprocess
 
 import skrt.image
 from skrt.structures import ROI, StructureSet
-from skrt.core import get_logger, Data, to_list, Defaults
+from skrt.core import fullpath, get_logger, Data, to_list, Defaults, PathData
 from skrt.dose import ImageOverlay, Dose
 from skrt.simulation import make_grid
 
@@ -106,7 +107,7 @@ class Registration(Data):
         self.capture_output = capture_output
 
         # Set up directory
-        self.path = path
+        self.path = fullpath(path)
         if not os.path.exists(path):
             os.makedirs(path)
         elif overwrite:
@@ -140,8 +141,8 @@ class Registration(Data):
         self.transformed_images = {}
         self.jacobians = {}
         self.deformation_fields = {}
-        if isinstance(pfiles, str):
-            pfiles = [pfiles]
+        if isinstance(pfiles, (str, Path)):
+            pfiles = [str(pfiles)]
         if pfiles is not None:
             self.add_pfiles(pfiles)
         else:
@@ -188,7 +189,7 @@ class Registration(Data):
             im = skrt.image.Image(im)
         path = getattr(self, f"{category}_path")
         if not os.path.exists(path) or force:
-            skrt.image.Image.write(im, path)
+            skrt.image.Image.write(im, path, verbose=(self.logger.level < 30))
         if 'grid' in category or 'mask' in category:
             setattr(self, f"{category}", skrt.image.Image(path))
         else:
@@ -221,8 +222,8 @@ class Registration(Data):
         for category in ["fixed", "moving"]:
             path = getattr(self, f"{category}_path")
             if not os.path.exists(path):
-                print(
-                    f"Warning: no {category} image found at {path}! "
+                self.logger.warning(
+                    f"No {category} image found at {path}! "
                     f"Make sure you run Registration.set_{category}_image"
                     " before running a registration."
                 )
@@ -254,9 +255,10 @@ class Registration(Data):
         # Infer name from parameter file name if name is None
         if name is None:
             if isinstance(pfile, dict):
-                print("If passing parameters from dict, <name> must be set.")
+                self.logger.warning(
+                        "If passing parameters from dict, <name> must be set.")
                 return
-            name = os.path.basename(pfile).replace(".txt", "")
+            name = os.path.basename(fullpath(pfile)).replace(".txt", "")
 
         # Check whether name already exists and add counter if so
         i = 1
@@ -322,7 +324,8 @@ class Registration(Data):
         if not filename.endswith(".txt"):
             filename += ".txt"
         if filename not in files:
-            print(f"Default file {name} not found. Available files:")
+            self.logger.warning(
+                    f"Default file {name} not found. Available files:")
             self.list_default_pfiles(self)
             return
         full_files = get_default_pfiles(False)
@@ -350,7 +353,8 @@ class Registration(Data):
         if not filename.endswith(".txt"):
             filename += ".txt"
         if filename not in files:
-            print(f"Default file {name} not found. Available files:")
+            self.logger.warning(
+                    f"Default file {name} not found. Available files:")
             self.list_default_pfiles(self)
             return
 
@@ -388,8 +392,8 @@ class Registration(Data):
         for step in steps:
             outdir = os.path.join(self.path, step)
             if not os.path.exists(outdir):
-                print(
-                    f"Warning: no output directory ({self.path}/{step}) "
+                self.logger.warning(
+                    f"No output directory ({self.path}/{step}) "
                     f"found for registration step {step} listed in "
                     f"{self.steps_file}. This step will be ignored."
                 )
@@ -397,8 +401,8 @@ class Registration(Data):
 
             pfile = os.path.join(outdir, "InputParameters.txt")
             if not os.path.exists(pfile):
-                print(
-                    f"Warning: no parameter file ({outdir}/InputParameters.txt) "
+                self.logger.warning(
+                    f"No parameter file ({outdir}/InputParameters.txt) "
                     f"found for registration step {step} listed in "
                     f"{self.steps_file}. This step will be ignored."
                 )
@@ -468,8 +472,8 @@ class Registration(Data):
         if use_previous_tfile and i > 0:
             prev_step = self.steps[i - 1]
             if prev_step not in self.tfiles:
-                print(
-                    f"Warning: previous step {prev_step} has not yet "
+                self.logger.warning(
+                    f"Previous step {prev_step} has not yet "
                     f"been performed! Input transform file for step {step}"
                     " will not be used."
                 )
@@ -583,8 +587,8 @@ class Registration(Data):
         # Check whether registration succeeded
         if code:
             logfile = os.path.join(self.outdirs[step], "elastix.log")
-            print(
-                f"Warning: registration step {step} failed! See "
+            self.logger.warning(
+                f"Registration step {step} failed! See "
                 f"{logfile} or run Registration.print_log({step}) for "
                 " more info."
             )
@@ -642,7 +646,8 @@ class Registration(Data):
         elif isinstance(to_transform, StructureSet):
             return self.transform_structure_set(to_transform, **kwargs)
         else:
-            print(f"Unrecognised transform input type {type(to_transform)}")
+            self.logger.warning(
+                    f"Unrecognised transform input type {type(to_transform)}")
 
     def transform_image(self, im, step=-1, outfile=None, params=None, rois=None,
             ):
@@ -689,7 +694,7 @@ class Registration(Data):
             im_path = im.path
         else:
             im_path = os.path.join(self._tmp_dir, "image.nii.gz")
-            im.write(im_path, verbose=False)
+            im.write(im_path, verbose=(self.logger.level < 30))
 
         # Transform the nifti file
         result_path = self.transform_data(im_path, step, params)
@@ -767,8 +772,8 @@ class Registration(Data):
             if os.path.exists(logfile):
                 os.remove(logfile)
             shutil.move(os.path.join(self._tmp_dir, "transformix.log"), self.path)
-            print(
-                f"Warning: image transformation failed! See "
+            self.logger.warning(
+                f"Image transformation failed! See "
                 f"{logfile} for more info."
             )
             return
@@ -824,7 +829,7 @@ class Registration(Data):
         else:
             ext = 'txt' if transform_points else 'nii.gz'
             roi_path = os.path.join(self._tmp_dir, f"{roi.name}.{ext}")
-            roi.write(roi_path, verbose=False)
+            roi.write(roi_path, verbose=(self.logger.level < 30))
 
         # Set default parameters
         default_params = {"ResampleInterpolator": '"FinalNearestNeighborInterpolator"'}
@@ -904,7 +909,7 @@ class Registration(Data):
 
         # Write structure set if outname is given
         if outfile is not None:
-            final.write(outfile)
+            final.write(outfile, verbose=(self.logger.level < 30))
             return
 
         # Otherwise, return structure set
@@ -1081,7 +1086,7 @@ class Registration(Data):
             if len(self.steps) == 1:
                 step = self.steps[0]
             else:
-                print(
+                self.logger.warning(
                     "This registration has more than one step. The step to "
                     "be manually adjusted must be specified when running "
                     "Registration.manually_adjust_transform()."
@@ -1091,13 +1096,14 @@ class Registration(Data):
 
         # Check registration has been run
         if not self.is_registered(step):
-            print(f"Registration for {step} has not yet been performed.")
+            self.logger.warning(
+                    f"Registration for {step} has not yet been performed.")
             return
 
         # Check the tfile contains a 3-parameter translation
         pars = read_parameters(self.tfiles[step])
         if pars["Transform"] != "TranslationTransform":
-            print(
+            self.logger.warning(
                 f"Can only manually adjust a translation step. Incorrect "
                 f"transform type for step {step}: {pars['Transform']}"
             )
@@ -1135,7 +1141,8 @@ class Registration(Data):
 
         step = self.get_step_name(step)
         if not self.is_registered(step):
-            print(f"Registration step {step} has not yet been performed.")
+            self.logger.warning(
+                    f"Registration step {step} has not yet been performed.")
             return
         return read_parameters(self.tfiles[step])
 
@@ -1366,7 +1373,7 @@ class Jacobian(ImageOverlay):
         self._default_colorbar_label = "Jacobian determinant"
         self._default_vmin = -1
         self._default_vmax = 2
-        self._default_opacity = 0.5
+        self._default_opacity = 0.8
         self.load()
         self.data = -self.data
 
@@ -1374,14 +1381,14 @@ class Jacobian(ImageOverlay):
         return ImageOverlay.view(self, kwarg_name="jacobian", **kwargs)
 
 
-class DeformationField:
+class DeformationField(PathData):
     """Class representing a vector field."""
 
     def __init__(self, path, image=None, **kwargs):
         """Load vector field."""
 
-        # Store path
-        self.path = path
+        # Perform base-class initialisation.
+        super().__init__(path)
 
         # Initialise own image object
         self._image = skrt.image.Image(path, **kwargs)
@@ -1389,6 +1396,18 @@ class DeformationField:
         # Assign an associated Image
         self.image = image
 
+        # Initialise store for displacement images
+        self.displacement_images = {"x": None, "y": None, "z": None, "3d": None}
+
+        # Set some default plotting values.
+        self._default_cmap = "seismic"
+        self._default_colorbar_label = "displacement (mm)"
+        self._3d_colorbar_label = "3D displacement magnitude (mm)"
+        self._default_opacity = 0.8
+        self._default_vmin = -15
+        self._default_vmax = 15
+        self._quiver_colorbar_label = "2D displacement magnitude (mm)"
+
     def load(self, force=False):
 
         self._image.load()
@@ -1418,15 +1437,61 @@ class DeformationField:
         y = y.T
         return x, y, df_x, df_y
 
+    def get_displacement_image(self, displacement="3d"):
+        """
+        Get skrt.dose.ImageOverlay object representing point displacements.
+
+        **Parameter:**
+
+        displacement : str, default="3d"
+            Point displacement to be represented:
+            - 'x': displacement along x-axis;
+            - 'y': displacement along y-axis;
+            - 'z': displacement along z-axis;
+            - '3d': magnitude of 3d displacement.
+        """
+
+        # Check that type of information requested is unknown.
+        if displacement not in self.displacement_images:
+            print(f"Displacement type not known: '{displacement}'")
+            print(f"Known displacement types: {list(self.displacement_images)}")
+            return
+
+        # Check if OverlayImage of displacement requested is already stored.
+        # If not, then create and store it.
+        if self.displacement_images[displacement] is None:
+            im_data = self._image.clone().get_data()
+            idx = "xyz".find(displacement)
+            if -1 == idx:
+                # Magnitudes of 3d displacements.
+                im_data = np.sqrt(np.sum(np.square(im_data), axis=idx))
+            else:
+                # Signed displacements along axis.
+                im_data = np.squeeze(im_data[:, :, :, idx])
+            im = ImageOverlay(im_data, affine=self._image.get_affine())
+            im._default_cmap = self._default_cmap
+            if "3d" == displacement:
+                im._default_colorbar_label = self._3d_colorbar_label
+            else:
+                im._default_colorbar_label = (
+                        f"{displacement}-{self._default_colorbar_label}")
+            im._default_opacity = self._default_opacity
+            im._default_vmin = self._default_vmin
+            im._default_vmax = self._default_vmax
+            self.displacement_images[displacement] = im
+
+        return self.displacement_images[displacement]
+
     def plot(
         self,
         view="x-y",
         sl=None,
         idx=None,
         pos=None,
-        plot_type="quiver",
+        df_plot_type="quiver",
         include_image=False,
-        spacing=30,
+        df_opacity=None,
+        df_spacing=30,
         ax=None,
         gs=None,
         figsize=None,
@@ -1435,36 +1500,108 @@ class DeformationField:
         show=True,
         save_as=None,
         scale_in_mm=True,
-        **kwargs
+        title=None,
+        colorbar=False,
+        no_xlabel=False,
+        no_ylabel=False,
+        no_xticks=False,
+        no_yticks=False,
+        no_xtick_labels=False,
+        no_ytick_labels=False,
+        annotate_slice=False,
+        major_ticks=None,
+        minor_ticks=None,
+        ticks_all_sides=False,
+        no_axis_labels=False,
+        mask=None,
+        mask_threshold=0.5,
+        masked=True,
+        invert_mask=False,
+        mask_color="black",
+        **mpl_kwargs
     ):
 
         # Set up axes
         self._image.set_ax(view, ax, gs, figsize, zoom)
         self.ax = self._image.ax
         self.fig = self._image.fig
+        xlim = None
+        ylim = None
 
         # Plot the underlying image
         if include_image and self.image is not None:
-            self.image.plot(view, ax=self.ax, show=False)
+            self.image.plot(view, sl=sl, idx=idx, pos=pos, ax=self.ax, gs=gs,
+                    show=False, title="", colorbar=max((colorbar - 1), 0),
+                    no_xlabel=no_xlabel, no_ylabel=no_ylabel,
+                    no_xticks=no_xticks, no_yticks=no_yticks,
+                    no_xtick_labels=no_xtick_labels,
+                    no_ytick_labels=no_ytick_labels,
+                    mask=mask, mask_threshold=mask_threshold,
+                    masked=masked, invert_mask=invert_mask,
+                    mask_color=mask_color)
+            xlim = self.ax.get_xlim()
+            ylim = self.ax.get_ylim()
 
         # Get spacing in each direction in number of voxels
-        spacing = self.convert_spacing(spacing, scale_in_mm)
+        df_spacing = self.convert_spacing(df_spacing, scale_in_mm)
 
         # Get vectors and positions on this slice
         data_slice = self.get_slice(view, sl=sl, idx=idx, pos=pos, 
                                     scale_in_mm=scale_in_mm)
 
+        # Define plot's pre-zoom aspect ratio and axis limits.
+        im_kwargs = self._image.get_mpl_kwargs(view, None, scale_in_mm)
+        xlim = xlim or im_kwargs["extent"][0: 2]
+        ylim = ylim or im_kwargs["extent"][2: 4]
+        aspect = im_kwargs["aspect"]
+
+        # Define plot opacity.
+        mpl_kwargs["alpha"] = df_opacity or mpl_kwargs.get(
+                "alpha", self._default_opacity)
+
+        # Extract kwargs for colour bar and label.
+        clb_kwargs = mpl_kwargs.pop("clb_kwargs", {})
+        clb_label_kwargs = mpl_kwargs.pop("clb_label_kwargs", {})
+
         # Create plot
-        if plot_type == "quiver":
-            self._plot_quiver(view, data_slice, spacing, **kwargs)
-        elif plot_type == "grid":
-            self._plot_grid(view, data_slice, spacing, **kwargs)
+        if df_plot_type == "quiver":
+            self._plot_quiver(view, data_slice, df_spacing, colorbar,
+                    clb_kwargs, clb_label_kwargs, mpl_kwargs)
+        elif df_plot_type == "grid":
+            self._plot_grid(view, data_slice, df_spacing, mpl_kwargs)
+        elif df_plot_type in ["x-displacement", "y-displacement",
+                "z-displacement", "3d-displacement"]:
+            self._plot_displacement(df_plot_type=df_plot_type,
+                    view=view, sl=sl, idx=idx, pos=pos,
+                    include_image=False, ax=self.ax, gs=gs, show=False,
+                    title="", colorbar=colorbar, no_xlabel=no_xlabel,
+                    no_ylabel=no_ylabel, no_xticks=no_xticks,
+                    no_yticks=no_yticks, no_xtick_labels=no_xtick_labels,
+                    no_ytick_labels=no_ytick_labels,
+                    annotate_slice=annotate_slice,
+                    major_ticks=major_ticks, minor_ticks=minor_ticks,
+                    ticks_all_sides=ticks_all_sides,
+                    no_axis_labels=no_axis_labels, mask=mask,
+                    mask_threshold=mask_threshold, masked=masked,
+                    invert_mask=invert_mask, mask_color=mask_color,
+                    mpl_kwargs = mpl_kwargs)
         else:
-            raise ValueError(f"Unrecognised plot type {plot_type}")
+            print(f"Unrecognised plot type '{df_plot_type}'")
+
+        # Set plot's pre-zoom aspect ratio and axis limits.
+        self.ax.set_aspect(aspect)
+        self.ax.set_xlim(xlim)
+        self.ax.set_ylim(ylim)
 
         # Label and zoom axes
         idx = self._image.get_idx(view, sl=sl, idx=idx, pos=pos)
-        self._image.label_ax(view, idx=idx, scale_in_mm=scale_in_mm, **kwargs)
+        self._image.label_ax(view, idx=idx, scale_in_mm=scale_in_mm,
+                title=title, no_xlabel=no_xlabel, no_ylabel=no_ylabel,
+                no_xticks=no_xticks, no_yticks=no_yticks,
+                no_xtick_labels=no_xtick_labels,
+                no_ytick_labels=no_ytick_labels, annotate_slice=annotate_slice,
+                major_ticks=major_ticks, ticks_all_sides=ticks_all_sides,
+                no_axis_labels=no_axis_labels, **mpl_kwargs)
         self._image.zoom_ax(view, zoom, zoom_centre)
 
         # Display image
@@ -1482,6 +1619,9 @@ class DeformationField:
         view, 
         data_slice,
         spacing,
+        colorbar=0,
+        clb_kwargs=None,
+        clb_label_kwargs=None,
         mpl_kwargs=None, 
     ):
         """Draw a quiver plot."""
@@ -1495,24 +1635,52 @@ class DeformationField:
         plot_y = y[:: spacing[y_ax], :: spacing[x_ax]]
 
         # Make plotting kwargs
+        mpl_kwargs = mpl_kwargs or {}
+        vmin = mpl_kwargs.pop("vmin", 0)
+        vmax = mpl_kwargs.pop("vmax",
+                self.get_displacement_image("3d").get_data().max() * 1.1)
         default_kwargs = {"cmap": "jet"}
-        if mpl_kwargs is not None:
-            default_kwargs.update(mpl_kwargs)
+        default_kwargs.update(mpl_kwargs)
+        default_dot_colour = matplotlib.cm.get_cmap(default_kwargs["cmap"])(0)
+        clb_kwargs = clb_kwargs or None
+        clb_label_kwargs = clb_label_kwargs or None
 
         # Plot arrows
         if arrows_x.any() or arrows_y.any():
-            M = np.hypot(arrows_x, arrows_y)
-            self.ax.quiver(
-                plot_x,
-                plot_y,
-                arrows_x,
-                arrows_y,
-                M,
-                **default_kwargs
-            )
+            if "color" in default_kwargs:
+                self.ax.quiver(
+                        plot_x,
+                        plot_y,
+                        arrows_x,
+                        arrows_y,
+                        **default_kwargs
+                        )
+            else:
+                M = np.hypot(arrows_x, arrows_y)
+                quiver = self.ax.quiver(
+                        plot_x,
+                        plot_y,
+                        arrows_x,
+                        arrows_y,
+                        M,
+                        **default_kwargs
+                        )
+                # Add colorbar
+                if colorbar > 0 and mpl_kwargs.get(
+                        "alpha", self._default_opacity) > 0:
+                    scalar_mappable = matplotlib.cm.ScalarMappable(
+                            norm=matplotlib.colors.Normalize(vmin, vmax),
+                            cmap=default_kwargs["cmap"])
+                    clb = self.fig.colorbar(scalar_mappable, ax=self.ax,
+                            **clb_kwargs)
+                    clb.set_label(self._quiver_colorbar_label,
+                            **clb_label_kwargs)
+                    clb.solids.set_edgecolor("face")
         else:
             # If arrow lengths are zero, plot dots
-            ax.scatter(plot_x, plot_y, c="navy", marker=".")
+            dot_colour = default_kwargs.get("color", default_dot_colour)
+            ax.scatter(plot_x, plot_y, c=dot_colour, marker=".")
+
 
     def _plot_grid(
         self, 
@@ -1535,12 +1703,92 @@ class DeformationField:
         if mpl_kwargs is not None:
             default_kwargs.update(mpl_kwargs)
 
+        # Disregard kwargs that aren't valid here,
+        # but are valid for other representations of deformation field.
+        for key in ["cmap", "vmin", "vmax"]:
+            default_kwargs.pop(key, None)
+
         # Plot gridlines
         for i in np.arange(0, x.shape[0], spacing[y_ax]):
             self.ax.plot(grid_x[i, :], grid_y[i, :], **default_kwargs)
         for j in np.arange(0, x.shape[1], spacing[x_ax]):
             self.ax.plot(grid_x[:, j], grid_y[:, j], **default_kwargs)
 
+    def _plot_displacement(self, df_plot_type="3d-displacement", **kwargs):
+        """
+        Plot displacements along an axis, or magnitudes of 3d displacements.
+
+        **Parameter:**
+
+        df_plot_type : str, default="3d-displacement"
+            Type of plot to produce for deformation field.
+            This function handles the plot types:
+            'x-displacement', 'y-displacement', 'z-displacement',
+            '3d-displacement'
+        """
+
+        if not df_plot_type.endswith("-displacement"):
+            print(f"df_plot_type not recognised: {df_plot_type}")
+            return
+
+        displacement = df_plot_type.split("-")[0]
+        im = self.get_displacement_image(displacement)
+        if im is not None:
+
+            # Set default values for vmin, vmax, cmap, alpha.
+            kwargs["mpl_kwargs"] = kwargs.get("mpl_kwargs", {})
+            vmax = max(abs(im._default_vmax), abs(im.get_data().max()),
+                    abs(im._default_vmin), abs(im.get_data().min()), 0)
+
+            mpl_kwargs = {"vmin": -vmax, "vmax": vmax,
+                    "cmap": im._default_cmap, "alpha": im._default_opacity}
+            for key, default_value in mpl_kwargs.items():
+                kwargs["mpl_kwargs"][key] = (
+                        kwargs["mpl_kwargs"].get(key, default_value))
+
+            # Disregard kwargs that aren't valid here,
+            # but are valid for other representations of deformation field.
+            for key in ["color"]:
+                kwargs["mpl_kwargs"].pop(key, None)
+
+            # Plot image.
+            im.plot(**kwargs)
+
+    def view(self, include_image=False, **kwargs):
+        """
+        View the deformation field.
+
+        **Parameters:**
+        
+        include_image : bool, default=True
+            If True, the image associated with the deformation field will
+            be displayed as underlay.
+
+        Any ``**kwargs`` will be passed to BetterViewer initialisation.
+        """
+
+        from skrt.better_viewer import BetterViewer
+        self.load()
+
+        # Ensure that df keyword isn't passed also via kwargs.
+        kwargs.pop("df", None)
+
+        # Define image for display: the image associated with
+        # the deformation field, or a dummy image of the same size.
+        if include_image and self.image is not None:
+            im = self.image
+        else:
+            im = skrt.image.Image(
+                    np.ones(self._image.get_data().shape[0: 3]) * 1e4,
+                    affine=self._image.get_affine())
+            im._default_cmap = self._default_cmap
+            im._default_colorbar_label = self._default_colorbar_label
+            im._default_vmin = self._default_vmin
+            im._default_vmax = self._default_vmax
+        
+        # Create viewer
+        return BetterViewer(im, df=self, **kwargs)
+    
     def convert_spacing(self, spacing, scale_in_mm):
         """Convert grid spacing in mm or voxels to list containing grid spacing 
         in voxels in each dimension."""
@@ -1567,7 +1815,7 @@ def set_elastix_dir(path):
 
     # Set directory
     global _ELASTIX_DIR
-    _ELASTIX_DIR = path
+    _ELASTIX_DIR = fullpath(path)
 
     # Find elastix exectuable
     global _ELASTIX
@@ -1693,7 +1941,7 @@ def shift_translation_parameters(infile, dx=0, dy=0, dz=0, outfile=None):
     pars = read_parameters(infile)
     init = pars["TransformParameters"]
     if pars["Transform"] != "TranslationTransform":
-        print(
+        self.logger.warning(
             f"Can only manually adjust a translation step. Incorrect "
             f"transform type: {pars['Transform']}"
         )
@@ -1708,12 +1956,15 @@ def get_default_pfiles(basename_only=True):
     import skrt
 
     rel_path = "data/elastix_parameter_files".split("/")
-    pdir = os.path.join(skrt.__path__[0], *rel_path)
+    pdir = fullpath(os.path.join(skrt.__path__[0], *rel_path))
     files = [file for file in os.listdir(pdir) if file.endswith(".txt")]
     if basename_only:
         return files
     return [os.path.join(pdir, file) for file in files]
 
+def get_default_pfiles_dir():
+    """Get path to directory containing default parameter files."""
+    return str(Path(get_default_pfiles(basename_only=False)[0]).parent)
 
 def get_jacobian_colormap(col_per_band=100, sat_values={0: 1, 1: 0.5, 2: 1}):
     '''
